@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2017, University of Amsterdam
+    Copyright (c)  2011-2018, University of Amsterdam
                               VU University Amsterdam
+			      CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -439,7 +440,6 @@ S__wait(IOSTREAM *s)
     if ( rc < 0 && errno == EINTR )
     { if ( PL_handle_signals() < 0 )
       { Sset_exception(s, PL_exception(0));
-	s->flags |= SIO_CLEARERR;
 	errno = EPLEXCEPTION;
 	return -1;
       }
@@ -524,7 +524,6 @@ S__flushbuf(IOSTREAM *s)
     { if ( errno == EINTR )
       { if ( PL_handle_signals() < 0 )
 	{ Sset_exception(s, PL_exception(0));
-	  s->flags |= SIO_CLEARERR;
 	  errno = EPLEXCEPTION;
 	} else
 	  goto retry;
@@ -720,14 +719,6 @@ update_linepos(IOSTREAM *s, int c)
 
 
 
-int
-S__fcheckpasteeof(IOSTREAM *s, int c)
-{ S__checkpasteeof(s, c);
-
-  return c;
-}
-
-
 int
 S__fupdatefilepos_getc(IOSTREAM *s, int c)
 { IOPOS *p = s->position;
@@ -750,7 +741,6 @@ S__updatefilepos(IOSTREAM *s, int c)
   { update_linepos(s, c);
     p->charno++;
   }
-  S__checkpasteeof(s,c);
 
   return c;
 }
@@ -860,7 +850,7 @@ reperror(int c, IOSTREAM *s)
     return c;
   }
 
-  Sseterr(s, SIO_FERR|SIO_CLEARERR, "Encoding cannot represent character");
+  Sseterr(s, SIO_FERR, "Encoding cannot represent character");
   return -1;
 }
 
@@ -1350,7 +1340,7 @@ Sread_pending(IOSTREAM *s, char *buf, size_t limit, int flags)
   { int c = S__fillbuf(s);
 
     if ( c < 0 )
-    { if ( (s->flags & SIO_FEOF) )
+    { if ( (s->flags & SIO_FEOF) && Sfpasteof(s) != TRUE )
 	return 0;
       return c;
     }
@@ -1453,8 +1443,7 @@ ScheckBOM(IOSTREAM *s)
       return 0;
 
     if ( S__fillbuf(s) == -1 )
-    { if ( s->limitp - s->bufp > 0 )
-	s->flags &= ~SIO_FEOF;
+    { s->flags &= ~SIO_FEOF;
       return 0;				/* empty stream */
     }
     s->bufp--;
@@ -1548,7 +1537,7 @@ Sfpasteof(IOSTREAM *s)
 
 
 #define SIO_ERROR_FLAGS (SIO_FEOF|SIO_WARN|SIO_FERR| \
-			 SIO_FEOF2|SIO_TIMEOUT|SIO_CLEARERR)
+			 SIO_FEOF2|SIO_TIMEOUT)
 
 void
 Sclearerr(IOSTREAM *s)
@@ -1568,7 +1557,7 @@ Sclearerr(IOSTREAM *s)
 int
 Sseterr(IOSTREAM *s, int flags, const char *message)
 { for(; s && s->magic == SIO_MAGIC; s = s->upstream )
-  { s->flags = (s->flags & ~(SIO_WARN|SIO_FERR|SIO_CLEARERR)) | flags;
+  { s->flags = (s->flags & ~(SIO_WARN|SIO_FERR)) | flags;
 
     if ( s->message )
     { free(s->message);
