@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1999-2016, University of Amsterdam
+    Copyright (c)  1999-2017, University of Amsterdam
                               VU University Amsterdam
     All rights reserved.
 
@@ -47,6 +47,10 @@
 #define SIG_ALERT  SIGUSR2
 #endif
 
+#define GCREQUEST_AGC   0x01		/* GD->thread.gc.requests */
+#define GCREQUEST_CGC   0x02
+#define GCREQUEST_ABORT 0x04
+
 typedef enum
 { LDATA_IDLE = 0,
   LDATA_SIGNALLED,
@@ -54,18 +58,19 @@ typedef enum
   LDATA_ANSWERED
 } ldata_status_t;
 
-#define THREAD_STATUS_INVALID(s) ((s) == PL_THREAD_UNUSED || \
-				  (s) == PL_THREAD_RESERVED)
+#define THREAD_STATUS_INVALID(s) ((int)(s) < (int)PL_THREAD_RUNNING)
 
 typedef enum
 { PL_THREAD_UNUSED = 0,			/* no thread on this slot */
+  PL_THREAD_RESERVED,			/* allocated but not yet created */
+  PL_THREAD_JOINED,			/* just joined */
+					/* normal states */
   PL_THREAD_RUNNING,			/* a normally running one */
   PL_THREAD_EXITED,			/* died with thread_exit/1 */
   PL_THREAD_SUCCEEDED,			/* finished with Yes */
   PL_THREAD_FAILED,			/* finished with No */
   PL_THREAD_EXCEPTION,			/* finished with exception */
   PL_THREAD_NOMEM,			/* couldn't start due no-memory */
-  PL_THREAD_RESERVED,			/* allocated but not yet created */
   PL_THREAD_CREATED,			/* just created */
 } thread_status;
 
@@ -84,11 +89,10 @@ typedef struct
 
 typedef struct _PL_thread_info_t
 { int		    pl_tid;		/* Prolog thread id */
-  size_t	    local_size;		/* Stack sizes */
-  size_t	    global_size;
-  size_t	    trail_size;
-  size_t	    stack_size;		/* system (C-) stack */
-  int		    (*cancel)(int id);	/* cancel function */
+  size_t	    stack_limit;	/* Stack sizes */
+  size_t	    table_space;	/* Max size for local tables */
+  size_t	    c_stack_size;	/* system (C-) stack */
+  rc_cancel	    (*cancel)(int id);	/* cancel function */
   unsigned short    open_count;		/* for PL_thread_detach_engine() */
   unsigned	    detached      : 1;	/* detached thread */
   unsigned	    debug         : 1;	/* thread can be debugged */
@@ -190,7 +194,7 @@ extern counting_mutex _PL_mutexes[];	/* Prolog mutexes */
 
 #define L_MISC		0
 #define L_ALLOC		1
-#define L_ATOM		2
+#define L_REHASH_ATOMS	2
 #define L_FLAG	        3
 #define L_FUNCTOR	4
 #define L_RECORD	5
@@ -207,14 +211,16 @@ extern counting_mutex _PL_mutexes[];	/* Prolog mutexes */
 #define L_OP	       16
 #define L_INIT	       17
 #define L_TERM	       18
-#define L_GC	       19
-#define L_AGC	       20
-#define L_FOREIGN      21
-#define L_OS	       22
-#define L_LOCALE       23
+#define L_FOREIGN      19
+#define L_OS	       20
+#define L_LOCALE       21
+#define L_SORTR        22
+#define L_UMUTEX       23
+#define L_INIT_ATOMS   24
+#define L_CGCGEN       25
 #ifdef __WINDOWS__
-#define L_DDE	       25
-#define L_CSTACK       26
+#define L_DDE	       26
+#define L_CSTACK       27
 #endif
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -372,7 +378,8 @@ int			PL_thread_raise(int tid, int sig);
 COMMON(void)		cleanupThreads(void);
 COMMON(intptr_t)	system_thread_id(PL_thread_info_t *info);
 COMMON(double)	        ThreadCPUTime(PL_local_data_t *ld, int which);
-
+COMMON(void)		get_current_timespec(struct timespec *time);
+COMMON(void)	        carry_timespec_nanos(struct timespec *time);
 
 		 /*******************************
 		 *	 GLOBAL GC SUPPORT	*
@@ -385,7 +392,7 @@ COMMON(void)	resumeThreads(void);
 COMMON(void)	markAtomsMessageQueues(void);
 COMMON(void)	markAtomsThreadMessageQueue(PL_local_data_t *ld);
 
-#define acquire_ldata(ld)	(LD->thread.info->access.ldata = (ld))
+#define acquire_ldata(info)	acquire_ldata__LD(info PASS_LD)
 #define release_ldata(ld)	(LD->thread.info->access.ldata = NULL)
 
 #else /*O_PLMT, end of threading-stuff */
@@ -426,6 +433,14 @@ COMMON(double)	        ThreadCPUTime(PL_local_data_t *ld, int which);
 		 *	       COMMON		*
 		 *******************************/
 
+typedef struct
+{ functor_t functor;			/* functor of property */
+  int (*function)();			/* function to generate */
+} tprop;
+
+COMMON(int)	get_prop_def(term_t t, atom_t expected,
+			     const tprop *list, const tprop **def);
+
 COMMON(void)	initPrologThreads(void);
 COMMON(int)	pl_atom_table_in_use(AtomTable atom_table);
 COMMON(int)	pl_atom_bucket_in_use(Atom *atom_bucket);
@@ -433,10 +448,13 @@ COMMON(Atom**)	pl_atom_buckets_in_use(void);
 COMMON(Definition*)	predicates_in_use(void);
 COMMON(int)	pl_functor_table_in_use(FunctorTable functor_table);
 COMMON(int)	pl_kvs_in_use(KVS kvs);
-COMMON(int)	pushPredicateAccess__LD(Definition def, gen_t gen ARG_LD);
+COMMON(void)	cgcActivatePredicate__LD(Definition def, gen_t gen ARG_LD);
+COMMON(gen_t)	pushPredicateAccess__LD(Definition def ARG_LD);
 COMMON(void)	popPredicateAccess__LD(Definition def ARG_LD);
 COMMON(size_t)	popNPredicateAccess__LD(size_t n ARG_LD);
 COMMON(void)	markAccessedPredicates(PL_local_data_t *ld);
-COMMON(void)    cgc_thread_stats(cgc_stats *stats ARG_LD);
+COMMON(int)     cgc_thread_stats(cgc_stats *stats ARG_LD);
+COMMON(int)	signalGCThread(int sig);
+COMMON(int)	isSignalledGCThread(int sig ARG_LD);
 
 #endif /*PL_THREAD_H_DEFINED*/
